'''Generic async AWS functions for SQS.'''

# pylint: disable=dangerous-default-value
# pylint: disable=too-many-arguments

import logging
from typing import Dict
from typing import List

from botocore.exceptions import ClientError

from aioradio.aws.utils import AwsServiceManager

LOG = logging.getLogger(__name__)
AWS_SERVICE = AwsServiceManager(service='sqs', regions=['us-east-1', 'us-east-2'])
SQS = AWS_SERVICE.service_dict


@AWS_SERVICE.active
async def create_queue(queue: str, region: str, attributes: Dict[str, str]) -> Dict[str, str]:
    '''Create SQS queue in region defined.'''

    return await SQS[region]['client'].create_queue(QueueName=queue, Attributes=attributes)


@AWS_SERVICE.active
async def get_messages(
        queue: str,
        region: str,
        wait_time: int=20,
        max_messages: int=10,
        visibility_timeout: int=30,
        attribute_names: List[str]=[]) -> List[dict]:
    """
    Get up to 10 messages from an SQS queue. Returns a list of dicts where each dict contains
    the message information, Here is an example of a message produce from an s3 -> sqs event:
    {
        'MessageId': '0050daf1-313b-4a5c-a4e7-8e5596085fa8',
        'ReceiptHandle': '<very-long-string>',
        'MD5OfBody': 'ec3212bbe0cf0239ba54eefd206338ef',
        'Body': '{
            "Records": [{
                "eventVersion": "2.1",
                "eventSource": "aws:s3",
                "awsRegion": "us-east-2",
                "eventTime": "2020-07-07T19:02:45.192Z",
                "eventName": "ObjectCreated:CompleteMultipartUpload",
                "userIdentity": {"principalId":"AWS:AIDATIJBHOZJSHFN3H2KY"},
                "requestParameters": {"sourceIPAddress":"52.249.199.100"},
                "responseElements": {"x-amz-request-id":"625BA0F414478E41",
                "x-amz-id-2": "<very-long-string>"},
                "s3": {
                    "s3SchemaVersion": "1.0",
                    "configurationId": "tf-s3-queue-20191002201742888700000006",
                    "bucket": {
                        "name": "nrccua-datalab-efi-input-sandbox.us-east-2",
                        "ownerIdentity": {"principalId":"A1MQ0EIGU3DVVT"},
                        "arn": "arn:aws:s3:::nrccua-datalab-efi-input-sandbox.us-east-2"
                    },
                    "object": {
                        "key": "XXXXXX/hello_world.txt",
                        "size":23, "eTag":"bccd05d8b202eba5e812bcf501c1682a-1",
                        "versionId": "ZLMCPDFMu6W865WWyfiVsaWZU8pJRUb3",
                        "sequencer":"005F04C6DA2DF81A97"
                    }
                }
            }]
        }'
    }
    """

    messages = []
    resp = await SQS[region]['client'].get_queue_url(QueueName=queue)
    queue_url = resp['QueueUrl']
    resp = await SQS[region]['client'].receive_message(
        QueueUrl=queue_url,
        WaitTimeSeconds=wait_time,
        MaxNumberOfMessages=max_messages,
        VisibilityTimeout=visibility_timeout,
        AttributeNames=attribute_names)
    if 'Messages' in resp:
        messages = resp['Messages']

    return messages


@AWS_SERVICE.active
async def send_messages(
        queue: str,
        region: str,
        entries: List[Dict[str, str]]) -> Dict[str, list]:
    '''
    Send up to 10 messages to an SQS queue. Each dict in entries must have
    the keys: Id and MessageBody populated. The returned data is a dict with two keys, either
    Successful or Failed, for example:
    {
        'Successful': [
            {
                'Id': 'string',
                'MessageId': 'string',
                'MD5OfMessageBody': 'string',
                'MD5OfMessageAttributes': 'string',
                'MD5OfMessageSystemAttributes': 'string',
                'SequenceNumber': 'string'
            },
        ],
        'Failed': [
            {
                'Id': 'string',
                'SenderFault': True|False,
                'Code': 'string',
                'Message': 'string'
            },
        ]
    }
    '''

    resp = await SQS[region]['client'].get_queue_url(QueueName=queue)
    queue_url = resp['QueueUrl']
    result = await SQS[region]['client'].send_message_batch(QueueUrl=queue_url, Entries=entries)

    return result


@AWS_SERVICE.active
async def delete_messages(
        queue: str,
        region: str,
        entries: List[Dict[str, str]]) -> Dict[str, list]:
    '''
    Delete up to 10 messages from an SQS queue. Each dict in entries must have
    the keys: Id and ReceiptHandle populated. The returned data is a dict with two keys, either
    Successful or Failed, for example:
    {
        'Successful': [
            {
                'Id': 'string'
            },
        ],
        'Failed': [
            {
                'Id': 'string',
                'SenderFault': True|False,
                'Code': 'string',
                'Message': 'string'
            },
        ]
    }
    '''

    resp = await SQS[region]['client'].get_queue_url(QueueName=queue)
    queue_url = resp['QueueUrl']
    result = await SQS[region]['client'].delete_message_batch(QueueUrl=queue_url, Entries=entries)

    return result


@AWS_SERVICE.active
async def purge_messages(queue: str, region: str) -> str:
    '''Purge messages from queue in region defined.'''

    error = ''
    try:
        resp = await SQS[region]['client'].get_queue_url(QueueName=queue)
        queue_url = resp['QueueUrl']
        await SQS[region]['client'].purge_queue(QueueUrl=queue_url)
    except ClientError as err:
        if err.response['Error']['Code'] == 'AWS.SimpleQueueService.PurgeQueueInProgress':
            error = err.response['Error']['Message']
        else:
            raise

    return error




'''Generic async utility functions.'''

# pylint: disable=broad-except
# pylint: disable=logging-fstring-interpolation
# pylint: disable=protected-access

import asyncio
import logging
from asyncio import sleep
from dataclasses import dataclass
from dataclasses import field
from time import time
from typing import List

import aioboto3
import aiobotocore
import aiojobs

LOG = logging.getLogger(__name__)


@dataclass
class AwsServiceManager:
    '''AWS Service Manager'''

    service: str
    module: str = 'aiobotocore'
    sleep_interval: int = 300
    regions: List[str] = field(default_factory=list)
    scheduler: aiojobs._scheduler = None

    def __post_init__(self):
        '''Post constructor.'''

        services = ['s3', 'sqs', 'secretsmanager', 'dynamodb']
        if self.service not in services:
            raise ValueError(f'service parameter must be one of the following: {services}')

        modules = ['aiobotocore', 'aioboto3']
        if self.module not in modules:
            raise ValueError(f'module parameter must be one of the following: {modules}')

        if self.regions:
            self.service_dict = {r: {'active': 0, 'client': None, 'resource': None, 'session': None, 'busy': True} for r in self.regions}
        else:
            self.service_dict = {'active': 0, 'client': None, 'resource': None, 'session': None, 'busy': True}

        try:
            loop = asyncio.get_event_loop()
            if loop and loop.is_running():
                asyncio.create_task(self.create_scheduler())
            else:
                loop.run_until_complete(self.create_scheduler())
        except RuntimeError:
            loop = asyncio.new_event_loop().run_until_complete(self.create_scheduler())

    def __del__(self):
        self.scheduler.close()

    async def create_scheduler(self):
        '''Schedule jobs.'''

        self.scheduler = await aiojobs.create_scheduler()
        if self.regions:
            for region in self.regions:
                await self.scheduler.spawn(self.aio_server(item='client', region=region))
                if self.module == 'aioboto3':
                    await self.scheduler.spawn(self.aio_server(item='resource', region=region))
        else:
            await self.scheduler.spawn(self.aio_server(item='client'))
            if self.module == 'aioboto3':
                await self.scheduler.spawn(self.aio_server(item='resource'))

    async def aio_server(self, item, region=None):
        '''Begin long running server establishing modules service_dict object.'''

        service_dict = self.service_dict if region is None else self.service_dict[region]
        await self.establish_client_resource(service_dict, item, region=region)

        while True:
            # sleep for defined interval
            await sleep(self.sleep_interval)

            # if all functions are inactive proceed to re-establishing AioSession client
            start = time()
            while service_dict['active'] and (time() - start) < 300:
                await sleep(0.001)

            await self.establish_client_resource(service_dict, item=item, region=region, reestablish=True)

    async def establish_client_resource(self, service_dict, item, region=None, reestablish=False):
        '''Establish the AioSession client or resource, then re-establish every self.sleep_interval seconds.'''

        kwargs = {'service_name': self.service, 'verify': False}
        if region is None:
            phrase = f're-establish {self.service}' if reestablish else f'establish {self.service}'
        else:
            kwargs['region_name'] = region
            phrase = f're-establish {self.service} {region}' if reestablish else f'establish {self.service} {region}'

        LOG.info(f'Atempting to {phrase} service object...')
        if reestablish:
            service_dict['busy'] = True
            await service_dict[item].__aexit__(None, None, None)

        if self.module == 'aiobotocore':
            service_dict['session'] = aiobotocore.get_session()
            service_dict[item] = await service_dict['session'].create_client(**kwargs).__aenter__()
        elif self.module == 'aioboto3':
            func = aioboto3.client if item == 'client' else aioboto3.resource
            service_dict[item] = await func(**kwargs).__aenter__()

        service_dict['busy'] = False
        LOG.info(f'Successfully {phrase} service object!')

    def get_region(self, args, kwargs):
        '''Attempt to detect the region from the kwargs or args.'''

        region = None

        if 'region' in kwargs:
            region = kwargs['region']
        else:
            for name in args:
                if name in self.regions:
                    region = name
                    break

        return region

    def active(self, func):
        '''Decorator to keep track of currently running functions, allowing the AioSession client
        to only be re-establish when the count is zero to avoid functions using a stale client.'''

        async def wrapper(*args, **kwargs):
            '''Decorator wrapper.'''

            obj = self.service_dict[self.get_region(args, kwargs)] if self.regions else self.service_dict

            # Make sure aiobotocore or aioboto3 isn't busy
            while obj['busy']:
                await sleep(0.01)

            error = None
            obj['active'] += 1
            try:
                result = await func(*args, **kwargs)
            except Exception as err:
                error = err
            finally:
                obj['active'] -= 1
                if error is not None:
                    raise error

            return result

        return wrapper



'''Pyodbc functions for connecting and send queries.'''

# pylint: disable=c-extension-no-member

import os
from typing import Any
from typing import List
from typing import Union

import pyodbc

# driver location varies based on OS.  add to this list if necessary...
UNIXODBC_DRIVER_PATHS = [
    '/usr/lib/libtdsodbc.so',
    '/usr/local/lib/libtdsodbc.so',
    '/usr/lib/x86_64-linux-gnu/odbc/libtdsodbc.so'
]


async def get_unixodbc_driver_path(paths) -> Union[str, None]:
    '''Check the file system for the unixodbc driver.'''

    driver_path = None
    for path in paths:
        if os.path.exists(path):
            driver_path = path
            break

    return driver_path


async def establish_pyodbc_connection(
        host: str,
        user: str,
        pwd: str,
        driver: str = None) -> pyodbc.Connection:
    '''Acquire and return pyodbc.Connection object else raise FileNotFoundError.'''

    if driver is None:
        verified_driver = await get_unixodbc_driver_path(UNIXODBC_DRIVER_PATHS)
    else:
        verified_driver = await get_unixodbc_driver_path([driver])

    if verified_driver is None:
        raise FileNotFoundError('Unable to locate unixodbc driver file: libtdsodbc.so')

    return pyodbc.connect(
        f'DRIVER={verified_driver};SERVER={host};PORT=1433;UID={user};PWD={pwd};TDS_Version=8.0')


async def pyodbc_query_fetchone(conn: pyodbc.Connection, query: str) -> List[Any]:
    '''Execute pyodbc query and fetchone, see https://github.com/mkleehammer/pyodbc/wiki/Cursor'''

    cursor = conn.cursor()
    result = cursor.execute(query).fetchone()

    return result


async def pyodbc_query_fetchall(conn: pyodbc.Connection, query: str) -> List[Any]:
    '''Execute pyodbc query and fetchone, see https://github.com/mkleehammer/pyodbc/wiki/Cursor'''

    cursor = conn.cursor()
    result = cursor.execute(query).fetchall()

    return result



'''Generic functions related to Jira.'''

from typing import Any
from typing import Dict

import httpx


async def post_jira_issue(url: str, jira_user: str, jira_token: str, payload: Dict[str, Any]):
    '''Post payload to create jira issue.'''

    headers = {'Content-Type': 'application/json'}
    auth = (jira_user, jira_token)
    async with httpx.AsyncClient() as client:
        return await client.post(url=url, json=payload, auth=auth, headers=headers)


async def get_jira_issue(url: str, jira_user: str, jira_token: str):
    '''Get Jira issue using jira_link built with the expected jira_id,
    an example: https://nrccua.atlassian.net/rest/api/2/issue/<jira_id>.'''

    headers = {'Content-Type': 'application/json'}
    auth = (jira_user, jira_token)
    async with httpx.AsyncClient() as client:
        return await client.get(url=url, auth=auth, headers=headers)


async def add_comment_to_jira(url: str, jira_user: str, jira_token: str, comment: str):
    '''Add Jira comment to an existing issue.'''

    if not url.endswith('comment'):
        msg = 'Check url value! ' \
              'Good example is https://nrccua.atlassian.net/rest/api/2/issue/<jira_id>/comment'
        raise ValueError(msg)

    return await post_jira_issue(
        url=url, payload={'body': comment}, jira_user=jira_user, jira_token=jira_token)



'''Generic logger logging to console or using json_log_formatter when logging in docker
for cleaner datadog logging.'''

# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes

import logging
import os
import sys
from datetime import datetime
from typing import List

from ddtrace import tracer
from pythonjsonlogger import jsonlogger


class CustomJsonFormatter(jsonlogger.JsonFormatter):
    '''Custom Json Formatter'''

    def add_fields(self, log_record, record, message_dict):
        '''normalize default set of fields.'''

        ddtags = self.get_ddtags(record, reserved=self._skip_fields)
        if ddtags:
            log_record['ddtags'] = ddtags
        super().add_fields(log_record, record, message_dict)
        if not log_record.get("timestamp"):
            now = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            log_record["timestamp"] = now
        log_record["level"] = log_record["level"].upper() if log_record.get("level") else record.levelname

    @staticmethod
    def get_ddtags(record, reserved):
        '''Add datadog tags in the format datadog expects.'''

        tags = {k: v for k, v in record.__dict__.items() if k not in reserved}
        return ','.join([f"{k}:{v}" for k, v in tags.items()])

class DatadogLogger():
    '''Custom class for JSON Formatter to include level and name.'''

    def __init__(
            self,
            main_logger='',
            datadog_loggers=List[str],
            log_level=logging.INFO,
            log_format="%(timestamp)d %(level)d %(name)d %(message)d",
            use_ddtrace=True
    ):

        self.logger = logging.getLogger(main_logger)
        self.logger.setLevel(log_level)
        self.log_level = log_level
        self.datadog_loggers = set(datadog_loggers) if datadog_loggers else []
        self.format = log_format

        if use_ddtrace:
            tracer.configure(
                hostname=os.getenv('DD_AGENT_HOST', ''),
                port=os.getenv('DD_TRACE_AGENT_PORT', '')
            )

        self.add_handlers()

    def add_handlers(self):
        '''Create log handlers.'''

        for name in self.datadog_loggers:
            logger = logging.getLogger(name)
            formatter = CustomJsonFormatter(self.format)
            handler = logging.StreamHandler(sys.stdout)
            handler.setLevel(self.log_level)
            handler.setFormatter(formatter)
            logger.addHandler(handler)
